import math
import random

import pyglet
from pyglet import resource

import base
import data
import monsters
import powerups
import blocks


class Portal(base.CollideSprite):
	collide_type = "portal"
	radius = 15
	
	parent_level = None
	to_level_index = 0
	
	def __init__(self, batch=None, level=None, index=0):
		img = data.load_anim('portal.png', 2, 8, 0.1, True)
		super(Portal, self).__init__(img, batch=batch)
		self.parent_level = level
		self.to_level_index = index
		
	def collide(self, into):
		if (into.collide_type == "player" and self.parent_level != None):
			self.parent_level.use_portal( self.to_level_index )
		return False

class Level(object):
	blocks = []
	npcs = []
	powerups = []
	projectiles = []
	portals = []
	
	background_batch = None
	corpse_batch = None
	sprite_batch = None
	projectile_batch = None
	
	background = None
	
	level_index = 0
	switch_to_level = None
	
	def __init__(self, difficulty=50):
		self.background_batch = pyglet.graphics.Batch()
		self.corpse_batch = pyglet.graphics.Batch()
		self.sprite_batch = pyglet.graphics.Batch()
		self.projectile_batch = pyglet.graphics.Batch()
		
		available_maps = ['map01.jpg', 'map03.jpg', 'map04.jpg']
		background_img = resource.image(data.filepath(random.choice(available_maps)))
		self.background = base.BaseSprite(background_img, self.background_batch)
		self.background.x, self.background.y = data.WINDOW_WIDTH /2, data.WINDOW_HEIGHT /2
		
		#these need to be initialised per instance, or they are static?
		self.blocks = []
		self.portals = []
		self.npcs = []
		self.powerups = []
		self.projectiles = []
		self.create_blocks()
		self.create_objects()
	
	def use_portal(self, to_level_index):
		# see if this is the same way we came in
		# create appropriate string
		self.switch_to_level = to_level_index
		#level_manager.switch_level( to_level_index )
	
	def create_blocks(self):
		top_block = blocks.Block( batch=self.background_batch, width=data.WINDOW_WIDTH, height=100 )
		bottom_block = blocks.Block( batch=self.background_batch, width=data.WINDOW_WIDTH, height=100 )
		left_block = blocks.Block( batch=self.background_batch, width=100, height=data.WINDOW_HEIGHT)
		right_block = blocks.Block( batch=self.background_batch, width=100, height=data.WINDOW_HEIGHT)
		
		top_block.moveto(data.WINDOW_WIDTH * 0.5, -50)
		bottom_block.moveto(data.WINDOW_WIDTH * 0.5, data.WINDOW_HEIGHT +50)
		left_block.moveto(-50, data.WINDOW_HEIGHT*0.5)
		right_block.moveto(data.WINDOW_WIDTH+50, data.WINDOW_HEIGHT*0.5)
		
		self.blocks.append(top_block)
		self.blocks.append(bottom_block)
		self.blocks.append(left_block)
		self.blocks.append(right_block)
	
	
	def create_objects(self, difficulty=10):
		#test_block = blocks.Block( batch=self.background_batch )
		#test_block.moveto(300, 300)
		#self.blocks.append(test_block)
		
		#num_zombies = int(difficulty / 6. + random.randint(0, 5))
		
		monster_lists = [ [monsters.Worm, 20], [monsters.Skeleton, 10], [monsters.Zombie, 3] ]
		
		for monster_list in monster_lists:
			for n in range(0, monster_list[1]):
				new_monster = monster_list[0](self.sprite_batch, self.corpse_batch)
				new_monster.moveto( random.randint(0, data.WINDOW_WIDTH), random.randint(0, data.WINDOW_HEIGHT) )
				self.npcs.append(new_monster)
	
	def add_portal(self, to_index):
		new_portal = Portal(self.sprite_batch, self, to_index)
		new_portal.moveto( random.randint(0, data.WINDOW_WIDTH), random.randint(0, data.WINDOW_HEIGHT) )
		self.portals.append(new_portal)
	
	def from_portal(self, from_level_index=None):
		if (from_level_index != None):
			for portal in self.portals:
				if (portal.to_level_index == from_level_index):
					#player comes out of this portal, position close to the centre
					c = data.WINDOW_WIDTH / 2 , data.WINDOW_HEIGHT / 2
					centre_angle = data.vector_to_rotation(c[0] - portal.x, c[1] - portal.y)
					offset = data.rotation_to_vector( centre_angle, 30 )
					
					data.player.moveto(portal.x + offset[0], portal.y + offset[1])
					data.player.dx += offset[0] * 8.
					data.player.dy += offset[1] * 8.
	
	def add_projectile(self, projectile_class):
		new_proj = projectile_class( batch=self.projectile_batch )
		self.projectiles.append( new_proj )
		new_proj.restore()
		return new_proj
	
	def draw(self):
		self.background_batch.draw()
		self.corpse_batch.draw()
		self.sprite_batch.draw()
		self.projectile_batch.draw()
		data.player.draw()
		
	def update(self, dt):
		if (self.switch_to_level != None):
			level_manager.switch_level( self.switch_to_level )
			self.switch_to_level = None
			
		lists = [ self.npcs, self.powerups, self.projectiles, self.portals ]
		
		for list in lists:
			for item in list:
				if (item.is_destroyed):
					list.remove(item)
					item.remove()
					item.delete()
				else:
					item.update(dt)
	
	def store(self):
		lists = [ self.blocks, self.npcs, self.powerups, self.projectiles, self.portals ]
		
		for list in lists:
			for item in list:
				item.store()
	
	def restore(self):
		lists = [ self.blocks, self.npcs, self.powerups, self.projectiles, self.portals ]
		
		for list in lists:
			for item in list:
				item.restore()

class LevelManager(object):
	
	levels = []
	current_level = None
	
	def __init__(self):
		self.reset()
	
	def reset(self):
		# create a bunch of levels
		for n in range(0, 5):
			new_level = self.create_level()
			new_level.level_index = n
		
		# now that we have all the levels, create portals in them to link them up
		portals_per_level = 3
		for portal in range(1, portals_per_level):
			for m in range(0, len(self.levels)):
				level = self.levels[m]
				if ( len(level.portals) < portals_per_level ):
					# see which levels are valid landing points still
					valid_targets = []
					for n in range(0, len(self.levels)):
						if ( n != m and len(self.levels[n].portals) < portals_per_level ):
							valid_targets.append( self.levels[n].level_index )
							
					if ( len(valid_targets) == 0 ):
						break
						
					random_target = random.choice(valid_targets)
					level.add_portal( random_target )
					self.levels[random_target].add_portal( level.level_index )
		
	def create_level(self, difficulty=50):
		new_level = Level(difficulty)
		self.levels.append(new_level)
		return new_level
	
	def switch_level(self, level_index=0):
		level_index = data.clip( level_index, 0, len(self.levels) )
		self.activate_level( self.levels[level_index] )
	
	def activate_level(self, new_level):
		player_from_level = None
		if (self.current_level != None):
			self.current_level.store()
			player_from_level = self.current_level.level_index
		
		self.current_level = new_level
		self.current_level.restore()
		self.current_level.from_portal( player_from_level )
	
		#print "CURRENT LEVEL INDEX IS: " + str(self.current_level.level_index)
	
			
	def update(self, dt):
		if ( self.current_level != None ):
			self.current_level.update(dt)
	
	def draw(self):
		if ( self.current_level != None ):
			self.current_level.draw()
	
level_manager = LevelManager()