import math
import random

import pyglet
from pyglet import resource

import base
import data
import levels

class Projectile(base.CollideSprite):
	collide_type = "projectile"
	
	parent_weapon = None
	
	mass = 0.5
	friction = 1
	radius = 3
	shape_group = 100

	impulse_speed = 10.
	accelerate_speed = 0.
	rotate_speed = 0.
	range = 1000.
	hit_damage = 5
	explode_delay = 0 # cause splash damage after timeout
	die_delay = 5 # fail after timeout
	
	age = 0
	
	start = []
	end = [0,0]
	
	def __init__(self, img, batch=None):
		super(Projectile, self).__init__(img, batch=batch)
		
	def update(self, dt):
		self.age += dt
		
		# change the projectile speed, if it accelerates
		if ( self.accelerate_speed > 0 ):
			# rotate towards our target
			targetrot = data.vector_to_rotation( self.end[0] - self.x, self.end[1] - self.y )
			rotdelta = data.clip(targetrot - self.rotation, -180, 180)
			
			rotspeed = self.rotate_speed * dt
			
			rotamount = rotdelta
			if ( rotdelta > 0 ):
				rotamount = min([rotdelta, rotspeed])
			if ( rotdelta < 0 ):
				rotamount = max([rotdelta, -rotspeed])
			
			self.rotation += rotamount
			
			#now we are aiming as best we can, apply correct vector
			accel_offset = data.rotation_to_vector( self.rotation, self.accelerate_speed * dt )
			
			self.dx = accel_offset[0]
			self.dy = accel_offset[1]
		
		super(Projectile, self).update(dt)
		
		# check for timeout
		if ( self.explode_delay > 0 and self.explode_delay < self.age ):
			self.explode()
			
		if ( self.die_delay > 0 and self.die_delay < self.age ):
			self.die()
		
		#check for any collisions
		travel_dist = data.dist(self.start[0], self.start[1], self.x, self.y)
		if (travel_dist > self.range):
			self.die()

	def explode(self, x=None, y=None):
		self.die()
		
	def fire(self, fromx, fromy, angle=0, tox=None, toy=None):
		# reset this projectile
		self.moveto(fromx, fromy)
		self.rotation = angle

		self.start = [fromx, fromy]
		
		self.end = data.rotation_to_vector( angle, self.range*2.0 )
		if (tox != None):
			self.end[0] = tox
		if (toy != None):
			self.end[1] = toy
			
		#apply impulse speed
		if ( self.impulse_speed > 0 ):
			initial_speed = data.rotation_to_vector( self.rotation, self.impulse_speed )
			self.dx = initial_speed[0]
			self.dy = initial_speed[1]
		
		#reset sprite animation

	def collide(self, into):
		# don't collide with whatever fired
		if ( self.parent_weapon != None and 
			self.parent_weapon.parent_character != None and 
			self.parent_weapon.parent_character == into ):
			return False
		elif ( into.collide_type == "explosion" or into.collide_type == "portal" ):
			return False
			
		explodex = self.x
		explodey = self.y
		if ( into.collide_type == "character" ):
			into.hurt( self.hit_damage )
			explodex = into.x + 0.5 * (self.x-into.x)
			explodey = into.y + 0.5 * (self.y-into.y)
			
		self.explode(explodex, explodey)
		return True

class Explosion(base.CollideSprite):
	collide_type = "explosion"
	
	parent_weapon = None
	
	radius = 25.
	shape_group = 101
	splash_damage = 10.

	die_delay = 1.
	
	age = 0
	
	def __init__(self, img=None, batch=None):
		if (img == None):
			#img = resource.image(data.filepath('fire.png'))
			img = data.load_anim('explosion.png', 2, 8, 0.05, False)
		super(Explosion, self).__init__(img, batch=batch)
		
	def update(self, dt):
		self.age += dt
		
		if ( self.die_delay > 0 and self.die_delay < self.age ):
			self.die()
		
	def explode(self, posx, posy):
		self.die()
		
	def collide(self, into):
		if ( self.parent_weapon != None and 
			self.parent_weapon.parent_character != None and 
			self.parent_weapon.parent_character == into ):
			return False
			
		if ( into.collide_type in ["character", "player"] ):
			# hurt based on distance
			char_distance = data.dist( into.x, into.y, self.x, self.y )
			distance_percent = 1.0 - (char_distance / self.radius)
			if ( distance_percent < 0 ):
				distance_percent = 0
				
			into.hurt( self.splash_damage * distance_percent )
			push_amount = 2.0 * self.splash_damage * distance_percent 
			into.dx += (into.x - self.x) * push_amount
			into.dy += (into.y - self.y) * push_amount
			
		return False

class NukeExplosion(Explosion):
	radius = 100.
	shape_group = 101
	splash_damage = 30.

	die_delay = 1.
	
	def __init__(self, batch=None):
		img = data.load_anim('explosion.png', 2, 8, 0.05, False)
		super(Explosion, self).__init__(img, batch=batch)
		self.scale = 10
		
		
class Bullet(Projectile):
	def __init__(self, batch=None):
		img = resource.image(data.filepath('bullet.png'))
		super(Bullet, self).__init__(img, batch=batch)
		self.reset()
		
	def reset(self):
		self.mass = 0.4
		self.impulse_speed = 10.


class Flame(Projectile):
	def __init__(self, batch=None):
		img = data.load_anim('flame.png', 2, 10, 0.05, False)
		super(Flame, self).__init__(img, batch=batch)
		self.reset()
		
	def reset(self):
		self.mass = 0.5
		self.impulse_speed = 3.
		self.friction = 2
		self.range = 100.
		self.hit_damage = 2.
		self.explode_delay = 0
		self.die_delay = 1

	def collide(self, into):
		if ( self.parent_weapon != None and 
			self.parent_weapon.parent_character != None and 
			self.parent_weapon.parent_character == into ):
			return False
		if ( into.collide_type == "character" ):
			into.hurt( self.hit_damage )
		return False

	def fire(self, fromx, fromy, angle=0, tox=None, toy=None):
		super(Flame, self).fire(fromx, fromy, angle, tox, toy)
		self.rotation = random.randint(0, 350)

class HomingRocket(Projectile):
	def __init__(self, img=None, batch=None):
		if (img == None):
			img = resource.image(data.filepath('rocket_small.png'))
		super(HomingRocket, self).__init__(img, batch=batch)
		self.reset()
		
	def reset(self):
		self.mass = 0.5
		self.impulse_speed = 15.
		self.accelerate_speed = 10.
		self.rotate_speed = 60.
		self.range = 800.
		self.hit_damage = 10.
		self.die_delay = 5 # last for 5 seconds
	
	def explode(self, posx=None, posy=None):
		if ( posx == None ):
			posx = self.x
		if ( posy == None ):
			posy = self.y
		
		#remove the rocket
		self.die()
		#but create a fireball
		explosion = levels.level_manager.current_level.add_projectile( Explosion )
		explosion.parent_weapon = self.parent_weapon
		explosion.moveto(posx, posy)
		explosion.rotation = self.rotation
	
	def collide(self, into):
		if ( into.collide_type == "explosion" ):
			# if we pass close enough to the explosion, then we should also explode
			distance = data.dist( into.x, into.y, self.x, self.y )
			distance_percent = 1.0 - (distance / into.radius)
			if ( distance_percent > 0.9 ):
				self.explode()
			
			return False
		return super(HomingRocket, self).collide(into)

class NukeRocket(HomingRocket):
	def __init__(self, batch=None):
		img = resource.image(data.filepath('rocket.png'))
		super(NukeRocket, self).__init__(img, batch=batch)
		self.reset()
		
	def reset(self):
		self.mass = 0.5
		self.impulse_speed = 10.
		self.accelerate_speed = 5.
		self.rotate_speed = 60.
		self.range = 800.
		self.hit_damage = 10.
		self.die_delay = 5
		self.explode_delay = 5

	def explode(self, posx=None, posy=None):
		if ( posx == None ):
			posx = self.x
		if ( posy == None ):
			posy = self.y
		
		#remove the rocket
		self.die()
		#but create a fireball
		explosion = levels.level_manager.current_level.add_projectile( NukeExplosion )
		explosion.parent_weapon = self.parent_weapon
		explosion.moveto(posx, posy)
		explosion.rotation = self.rotation
		
# monster projectiles
class ChewProjectile(Projectile):
	def __init__(self, batch=None):
		img = resource.image(data.filepath('none.png'))
		super(ChewProjectile, self).__init__(img, batch=batch)
		self.reset()
		
	def reset(self):
		self.mass = 0.5
		self.impulse_speed = 2.
		self.range = 25.
		self.hit_damage = 3.
		self.die_delay = 1